@@ -61,7 +61,6 @@ SOFTWARE.
 #include <X11/extensions/XInput.h>
 #include <X11/extensions/extutil.h>
 #include "XIint.h"
-#include <limits.h>
 
 XFeedbackState *
 XGetFeedbackControl(
@@ -69,6 +68,8 @@ XGetFeedbackControl(
     XDevice		*dev,
     int			*num_feedbacks)
 {
+    int size = 0;
+    int nbytes, i;
     XFeedbackState *Feedback = NULL;
     XFeedbackState *Sav = NULL;
     xFeedbackState *f = NULL;
@@ -90,28 +91,17 @@ XGetFeedbackControl(
 	goto out;
 
     if (rep.length > 0) {
-	unsigned long nbytes;
-	size_t size = 0;
-	int i;
-
 	*num_feedbacks = rep.num_feedbacks;
-
-	if (rep.length < (INT_MAX >> 2)) {
-	    nbytes = rep.length << 2;
-	    f = Xmalloc(nbytes);
-	}
+	nbytes = (long)rep.length << 2;
+	f = (xFeedbackState *) Xmalloc((unsigned)nbytes);
 	if (!f) {
-	    _XEatDataWords(dpy, rep.length);
+	    _XEatData(dpy, (unsigned long)nbytes);
 	    goto out;
 	}
 	sav = f;
 	_XRead(dpy, (char *)f, nbytes);
 
 	for (i = 0; i < *num_feedbacks; i++) {
-	    if (f->length > nbytes)
-		goto out;
-	    nbytes -= f->length;
-
 	    switch (f->class) {
 	    case KbdFeedbackClass:
 		size += sizeof(XKbdFeedbackState);
@@ -126,8 +116,6 @@ XGetFeedbackControl(
 	    {
 		xStringFeedbackState *strf = (xStringFeedbackState *) f;
 
-		if (strf->num_syms_supported >= (INT_MAX / sizeof(KeySym)))
-		    goto out;
 		size += sizeof(XStringFeedbackState) +
 		    (strf->num_syms_supported * sizeof(KeySym));
 	    }
@@ -142,12 +130,10 @@ XGetFeedbackControl(
 		size += f->length;
 		break;
 	    }
-	    if (size > INT_MAX)
-		goto out;
 	    f = (xFeedbackState *) ((char *)f + f->length);
 	}
 
-	Feedback = Xmalloc(size);
+	Feedback = (XFeedbackState *) Xmalloc((unsigned)size);
 	if (!Feedback)
 	    goto out;
 
@@ -195,18 +181,18 @@ XGetFeedbackControl(
 	    }
 	    case IntegerFeedbackClass:
 	    {
-		xIntegerFeedbackState *i;
+		xIntegerFeedbackState *ifs;
 		XIntegerFeedbackState *I;
 
-		i = (xIntegerFeedbackState *) f;
+		ifs = (xIntegerFeedbackState *) f;
 		I = (XIntegerFeedbackState *) Feedback;
 
-		I->class = i->class;
+		I->class = ifs->class;
 		I->length = sizeof(XIntegerFeedbackState);
-		I->id = i->id;
-		I->resolution = i->resolution;
-		I->minVal = i->min_value;
-		I->maxVal = i->max_value;
+		I->id = ifs->id;
+		I->resolution = ifs->resolution;
+		I->minVal = ifs->min_value;
+		I->maxVal = ifs->max_value;
 		break;
 	    }
 	    case StringFeedbackClass:
